/****
 * Name: compiler.cpp
 *
 * Details:
 * Called mainly by the parser to keep track of all allocations for
 * obvious reasons :-).
 * Also generates a PD array of type ProcDesc which is later used by
 * the compiler when interpreting or generating code.
 *
 * Bugs:
 ****/

#include <stdio.h>
#include <string.h>

#include "compiler.h"
#include "procdesc.h"

Compiler::Compiler( Machine *emc, int max )
{

   mc = emc;
   maxlevel = max;

   PDI = new (int[max]);
   CV = new (Decl (*[max]));

   for( int i = 0; i < max; i++ ) {
      CV[i] = 0;
      PDI[i] = 0;
   }


   blocklevel = 0;
}


Decl *Compiler::FindDecl( char *id )
{
   Decl *declaration = 0;
   int localblocklevel;

   localblocklevel = blocklevel;

   while( localblocklevel >= 0 ) {

      if( CV[localblocklevel] ) {

         if( declaration = CV[localblocklevel]->FindName( id ) ) break;

      }

      localblocklevel--;


   }

   return declaration;

}

int Compiler::NewVar( char *id )
{
   Decl *prev;
   Decl *newvar;

   int enter = 0;

   prev = FindDecl( id );

   if( prev ) if( prev->bllevel < blocklevel ) enter = 1;

   if( enter || !prev ) {

      newvar = new Decl( id, TYPE_VAR, mc->PD[PDI[blocklevel-1]]->ablen, blocklevel );
      mc->PD[PDI[blocklevel-1]]->ablen++;
      newvar->next = CV[blocklevel];
      CV[blocklevel] = newvar;

      return 1;

   }

   return 0;
}


int Compiler::NewProc( Decl *dproc )
{
   Decl *prev;
   ProcDesc *newproc;

   int enter = 0;

   prev = FindDecl( dproc->name );

   if( prev ) if( prev->bllevel < blocklevel ) enter == 1;

   if( enter || !prev ) {

      dproc->nr = mc->nextPD;
      dproc->next = CV[blocklevel];
      CV[blocklevel] = dproc;

      newproc = new ProcDesc;

      if( mc->CreateProc( newproc ) ) {

         PDI[blocklevel] = dproc->nr;
         newproc->blocklevel = blocklevel;
         strcpy( newproc->name, dproc->name );
         blocklevel++;

         return 1;
      }

   }

   return 0;
}

void Compiler::LeaveProc( void )
{

   CV[blocklevel] = 0;
   blocklevel--;

}

