/****
 * Name: machine.cpp
 *
 * Details:
 * The virtual machine.
 * Mainly used by the interpreter. Contains all normal machine stuff, like
 * the debugger etc. Also takes care of code generation if needed.
 * Note instbase.h refs!
 *
 * Bugs:
 * - Much static code which could be compiled and put in a separate link library.
 *   However, this approach eliminates the need for calling the linker at all.
 * - The Call implementation is still sloppy. The Owe Context Vector stuff is
 *   not added yet. Implementing this would make function call run in linear time (!).
 * - should redesign the machine to look more like a 68k based thing, not a single
 *   register hybrid. Should be a lot faster :-)
 * - register allocation scheme not implemented.
 ****/

#include <stdio.h>
#include <string.h>

#include "inst.h"
#include "procdesc.h"


Defs::Defs( Machine *emc )
{
   mc = emc;
}

int Defs::Execute( void )
{
   return 1;
}

int Defs::GenCode( FILE *outf )
{


   fprintf( outf, "\n" );

   fprintf( outf, "Init:\n" );
   fprintf( outf, "         move.l   $4.w,a6\n" );
   fprintf( outf, "         move.l   a6,_ExecBase\n" );
   fprintf( outf, "         move.l   #0,d0\n" );
   fprintf( outf, "         lea      DosName(pc),a1\n" );
   fprintf( outf, "         jsr      _LVOOpenLibrary(a6)\n" );
   fprintf( outf, "         move.l   d0,_DosBase\n" );
   fprintf( outf, "         beq      Cleanup\n" );

   fprintf( outf, "         move.l   #40000,d0\n" );
   fprintf( outf, "         move.l   #MEMF_CLEAR,d1\n" );
   fprintf( outf, "         jsr      _LVOAllocMem(a6)\n" );
   fprintf( outf, "         move.l   d0,Stack\n" );
   fprintf( outf, "         beq      Cleanup\n" );

   fprintf( outf, "         move.l   _DosBase(pc),a6\n" );
   fprintf( outf, "         jsr      _LVOInput(a6)\n" );
   fprintf( outf, "         move.l   d0,_Input\n" );
   fprintf( outf, "         jsr      _LVOOutput(a6)\n" );
   fprintf( outf, "         move.l   d0,_Output\n" );

   fprintf( outf, "         rts\n" );

   fprintf( outf, "\n" );
   fprintf( outf, "Cleanup:\n" );
   fprintf( outf, "         move.l   _ExecBase(pc),a6\n" );
   fprintf( outf, "         move.l   Stack,d0\n" );
   fprintf( outf, "         beq.b    .cl2\n" );
   fprintf( outf, "         move.l   d0,a1\n" );
   fprintf( outf, "         move.l   #40000,d0\n" );
   fprintf( outf, "         jsr      _LVOFreeMem(a6)\n" );
   fprintf( outf, ".cl2:    move.l   _DosBase(pc),d0\n" );
   fprintf( outf, "         beq.b    .cl3\n" );
   fprintf( outf, "         move.l   d0,a1\n" );
   fprintf( outf, "         jsr      _LVOCloseLibrary(a6)\n" );
   fprintf( outf, ".cl3:    move.l   SaveSP,a7\n" );
   fprintf( outf, "         rts\n" );


   // d0 = proindex

   fprintf( outf, "\n" );
   fprintf( outf, "SetCall:" );
   fprintf( outf, "         move.l   d0,d7\n" );
   fprintf( outf, "         lsl.l    #3,d0\n" );
   fprintf( outf, "         lea      PD(pc),a0\n" );
   fprintf( outf, "         add.l    d0,a0\n" );               // a0 = mc->PD[proindex] (proc)
   fprintf( outf, "         move.l   mc_cur(pc),d1\n" );       // d1 = mc_cur           (dynlink)
   fprintf( outf, "         move.l   mc_free(pc),mc_cur\n" );  // cur = free /Push
   fprintf( outf, "         move.l   0(a0),d0\n" );
   fprintf( outf, "         add.l    d0,mc_free\n" );          // free = free + size /Push

   fprintf( outf, "         move.l   mc_cur(pc),d2\n" );
   fprintf( outf, "         move.l   Stack(pc),a1\n" );
   fprintf( outf, "         add.l    d2,a1\n" );               // a1 = mc->GetCur()
   fprintf( outf, "         move.l   d1,OFF_DynLink(a1)\n" );  // mc->SetData(...DynLink...)

   fprintf( outf, "         move.l   4(a0),d3\n" );            // d3 = proc->blocklevel
   fprintf( outf, "         move.l   mc_bll(pc),d4\n" );
   fprintf( outf, "         addq.l   #1,d4\n" );               // d4 = mc->bll+1
   fprintf( outf, "         cmp.l    d3,d4\n" );
   fprintf( outf, "         bne.b    .sigh\n" );               // proc->blocklevel == mc->bll+1 ?

   fprintf( outf, "         move.l   d1,OFF_StatLink(a1)\n" ); // mc->SetData(...StatLink...)
   fprintf( outf, "         bra.b    .done\n" );

   fprintf( outf, ".sigh:   subq.l   #1,d4\n" );               // d4 = mc->bll
   fprintf( outf, "         sub.l    d3,d4\n" );               // d4 = mc->bll - proc->blocklevel (slsteps)

   fprintf( outf, ".loop:   move.l   Stack(pc),a2\n" );
   fprintf( outf, "         add.l    d1,a2\n" );               // a2 = Stack+abptr
   fprintf( outf, "         move.l   OFF_StatLink(a2),d1\n" );
   fprintf( outf, "         dbra     d4,.loop\n" );

   fprintf( outf, "         move.l   d1,OFF_StatLink(a1)\n" ); // mc->SetData(...StatLink...)

   fprintf( outf, ".done:   move.l   d3,mc_bll\n" );           // mc->bll = proc->blocklevel

   fprintf( outf, "         rts\n" );

   fprintf( outf, "\n" );
   fprintf( outf, "ClearCall:" );
   fprintf( outf, "         move.l   mc_free(pc),d0\n" );
   fprintf( outf, "         sub.l    mc_cur(pc),d0\n" );       // popsize

   fprintf( outf, "         move.l   mc_cur(pc),d1\n" );
   fprintf( outf, "         move.l   Stack(pc),a1\n" );
   fprintf( outf, "         add.l    d1,a1\n" );
   fprintf( outf, "         move.l   OFF_DynLink(a1),d1\n" );
   fprintf( outf, "         move.l   d1,mc_cur\n" );           // mc->SetCur(...)

   fprintf( outf, "         sub.l    d0,mc_free\n" );          // mc->Pop(...)
   fprintf( outf, "         rts\n" );


   fprintf( outf, "\n" );
   fprintf( outf, "OutChar:" );
   fprintf( outf, "         move.b   d0,(a3)+\n" );
   fprintf( outf, "         rts\n" );


   // d0 = sldist
   // returns d0 = correctblock offset

   fprintf( outf, "\n" );
   fprintf( outf, "GetBlock:\n" );
   fprintf( outf, "         move.l   mc_cur(pc),d1\n" );
   fprintf( outf, "         tst.l    d0\n" );
   fprintf( outf, "         bmi.b    .no\n" );

   fprintf( outf, ".hey:    move.l   Stack(pc),a1\n" );
   fprintf( outf, "         add.l    d1,a1\n" );
   fprintf( outf, "         move.l   OFF_StatLink(a1),d1\n" );
   fprintf( outf, "         dbra     d0,.hey\n" );

   fprintf( outf, ".no:     move.l   d1,d0\n" );
   fprintf( outf, "         rts\n" );


   // d0 = correctblock offset, d1 = reladr
   // returns d0 = value

   fprintf( outf, "\n" );
   fprintf( outf, "GetVar:\n" );
   fprintf( outf, "         move.l   Stack(pc),a1\n" );
   fprintf( outf, "         add.l    d0,a1\n" );
   fprintf( outf, "         move.l   (a1,d1.l),d0\n" );
   fprintf( outf, "         rts\n" );


   fprintf( outf, "\n" );
   fprintf( outf, "         cnop     0,4\n\n" );
   fprintf( outf, "SaveSP:           dc.l     0\n" );
   fprintf( outf, "_DosBase:         dc.l     0\n" );
   fprintf( outf, "_Input:           dc.l     0\n" );
   fprintf( outf, "_Output:          dc.l     0\n" );
   fprintf( outf, "_ExecBase:        dc.l     0\n" );
   fprintf( outf, "Stack:            dc.l     0\n" );
   fprintf( outf, "mc_free:          dc.l     3*4\n" );
   fprintf( outf, "mc_cur:           dc.l     0\n" );
   fprintf( outf, "mc_bll:           dc.l     0\n" );
   fprintf( outf, "mc_val:           dc.l     0\n" );
   fprintf( outf, "OutArg:           dc.l     0\n" );
   fprintf( outf, "mc_op:            dc.b     0\n" );
   fprintf( outf, "InBuf:            dcb.b    12,0\n" );
   fprintf( outf, "OutBuf:           dcb.b    80,0\n" );
   fprintf( outf, "OutString:        dc.b     \"Output: %%ld\",10,0\n" );
   fprintf( outf, "InBufAsk:         dc.b     \"Input: \",0" );

   fprintf( outf, "\n" );
   fprintf( outf, "DosName:          dc.b     \"dos.library\"\n" );

   fprintf( outf, "\n" );
   fprintf( outf, "         cnop     0,4\n\n" );
   fprintf( outf, "PD:\n" );
   for( int i = 0; i < mc->nextPD; i++ )
      fprintf( outf, "         dc.l     %d,%d\n", (mc->PD[i]->ablen)<<2, mc->PD[i]->blocklevel );

   return 1;
}


void Defs::Dump( void )
{
   printf( "Definitions (dummy)\n" );
}

FProgram::FProgram( Machine *emc )
{
   mc = emc;
}

int FProgram::Execute( void )
{
   return 1;
}

int FProgram::GenCode( FILE *outf )
{
   fprintf( outf, " MACHINE MC68020\n" );
   fprintf( outf, "_LVOOpenLibrary   equ      -552\n" );
   fprintf( outf, "_LVOCloseLibrary  equ      -414\n" );
   fprintf( outf, "_LVOAllocMem      equ      -198\n" );
   fprintf( outf, "_LVOFreeMem       equ      -210\n" );
   fprintf( outf, "_LVORawDoFmt      equ      -522\n" );
   fprintf( outf, "_LVORead          equ      -42\n" );
   fprintf( outf, "_LVOInput         equ      -54\n" );
   fprintf( outf, "_LVOOutput        equ      -60\n" );
   fprintf( outf, "_LVOFlush         equ      -360\n" );
   fprintf( outf, "_LVOPutStr        equ      -948\n" );
   fprintf( outf, "MEMF_CLEAR        equ      1<<16\n" );
   fprintf( outf, "OFF_DynLink       equ      0\n" );
   fprintf( outf, "OFF_StatLink      equ      4\n" );
   fprintf( outf, "OFF_RetAddr       equ      8\n" );


   fprintf( outf, "\n" );
   fprintf( outf, "         move.l   a7,SaveSP\n" );
   fprintf( outf, "         bsr      Init\n" );

   return 1;
}


void FProgram::Dump( void )
{
   printf( "Program startup (dummy)\n" );
}



FCall::FCall( Machine *emc, int eproindex )
{
   mc = emc;
   proindex = eproindex;
}

int FCall::Execute( void )
{

   ProcDesc *proc;
   int dynlink;

   dynlink = mc->GetCur();
   proc = mc->PD[proindex];

   if( !mc->Push( proc->ablen ) ) return 0;

   mc->SetData( mc->GetCur() + OFF_DynamicLink, dynlink );
   mc->SetData( mc->GetCur() + OFF_RetAddr, mc->GetIP() );

   if( proc->blocklevel == mc->bll + 1 )

      mc->SetData( mc->GetCur() + OFF_StaticLink, dynlink );

   else {

      if( proc->blocklevel <= mc->bll ) {

         int i;
         int slsteps;
         int abptr;

         abptr = dynlink;
         slsteps = mc->bll - proc->blocklevel;

         for( i = slsteps; i >= 0; i-- )
            abptr = mc->GetData( abptr + OFF_StaticLink );

         mc->SetData( mc->GetCur() + OFF_StaticLink, abptr );

      } else {
         puts( "Illegal procedure call!" );
         return 0;
      }


   }

   mc->bll = proc->blocklevel;
   mc->SetIP( proc->startinst );

   return 1;

}

int FCall::GenCode( FILE *outf )
{
   fprintf( outf, "         move.l   #%d,d0\n", proindex );
   fprintf( outf, "         bsr      SetCall\n" );
   fprintf( outf, "         bsr      lab_%d\n", proindex );
   return 1;
}


void FCall::Dump( void )
{
   printf( "Call(%d) %s\n", proindex, mc->PD[proindex]->name );
}


GLabel::GLabel( Machine *emc, int elabno )
{
   mc = emc;
   labno = elabno;
}

int GLabel::Execute( void )
{
   return 1;
}


int GLabel::GenCode( FILE *outf )
{
   fprintf( outf, "\n\nlab_%d:\n", labno );
   return 1;
}

void GLabel::Dump( void )
{
   printf( "GLabel = lab_%d (dummy)\n", labno );
}


LLabel::LLabel( Machine *emc, int ellabno )
{
   mc = emc;
   llabno = ellabno;
}

int LLabel::Execute( void )
{
   return 1;
}


int LLabel::GenCode( FILE *outf )
{
   fprintf( outf, "\n.llab_%d:\n", llabno );
   return 1;
}

void LLabel::Dump( void )
{
   printf( "LLabel = .llab_%d (dummy)\n", llabno );
}


FNumber::FNumber( Machine *emc, int eval )
{
   mc = emc;
   val = eval;
}

int FNumber::Execute( void )
{

   switch( mc->op ) {

   case 'A': case 'I':
      mc->val = val;
      break;

   case '+':
      mc->val+=val;
      break;

   case '-':
      mc->val-=val;
      break;

   case '*':
      mc->val*=val;
      break;

   case '/':
      mc->val/=val;
      break;

   default:
      puts( "Arithmetic error!" );
      return 0;
      break;

   }

//   printf( " evaluation: %d\n", mc->val );

   return 1;
}

int FNumber::GenCode( FILE *outf )
{

   switch( ((Operator *)(mc->IA[mc->currIA-1]))->operation ) {

   case 'A': case 'I':
      fprintf( outf, "         move.l   #%d,mc_val\n", val );
      break;

   case '+':
      fprintf( outf, "         add.l    #%d,mc_val\n", val );
      break;

   case '-':
      fprintf( outf, "         sub.l    #%d,mc_val\n", val );
      break;

   case '*':
      fprintf( outf, "         move.l   mc_val(pc),d0\n" );
      fprintf( outf, "         muls.l   #%d,d0\n", val );
      fprintf( outf, "         move.l   d0,mc_val\n" );
      break;

   case '/':
      fprintf( outf, "         move.l   mc_val(pc),d0\n" );
      fprintf( outf, "         divs.l   #%d,d0\n", val );
      fprintf( outf, "         move.l   d0,mc_val\n" );
      break;


   }

   return 1;
}

void FNumber::Dump( void )
{
   printf( "Number( %d )\n", val );
}


Var::Var( Machine *emc, int esldist, int ereladr, char *ename )
{
   mc = emc;

   sldist = esldist;
   reladr = ereladr;
   strcpy( name, ename );

}

int Var::Execute( void )
{
   int correctblock;
   int i;

   correctblock = mc->GetCur();
   i = sldist;

   if( i )  {

      while( i > 0 ) {

         correctblock = mc->GetData( correctblock + OFF_StaticLink );
         i--;

      }

   }

   // fuzzy...
   if( correctblock >= 0 ) {

      int value;

      value = mc->GetData( correctblock + reladr );

      switch( mc->op ) {



      case 'A': case 'I':
         mc->val = value;
         break;

      case '+':
         mc->val+=value;
         break;

      case '-':
         mc->val-=value;
         break;

      case '*':
         mc->val*=value;
         break;

      case '/':
         mc->val/=value;
         break;

      case '>':
         mc->SetData( correctblock+reladr, mc->val );
         break;

      case '?':
         printf( "Input for variable %s: ", name );
         scanf( "%d", &value );
         mc->SetData( correctblock+reladr, value );
         break;

      case '!':
         printf( "Output for variable %s: %d\n", name, value );
         break;

      }


//      printf( " evaluation: %d\n", mc->val );
      return 1;

   }


   return 0;
}

int Var::GenCode( FILE *outf )
{
   fprintf( outf, "         move.l   #%d,d0\n", sldist-1 );
   fprintf( outf, "         bsr      GetBlock\n" );

   switch( ((Operator *)(mc->IA[mc->currIA-1]))->operation ) {

   case 'A': case 'I':
      fprintf( outf, "         move.l   #%d,d1\n", (reladr<<2) );
      fprintf( outf, "         bsr      GetVar\n" );
      fprintf( outf, "         move.l   d0,mc_val\n" );
      break;

   case '+':
      fprintf( outf, "         move.l   #%d,d1\n", (reladr<<2) );
      fprintf( outf, "         bsr      GetVar\n" );
      fprintf( outf, "         add.l    d0,mc_val\n" );
      break;

   case '-':
      fprintf( outf, "         move.l   #%d,d1\n", (reladr<<2) );
      fprintf( outf, "         bsr      GetVar\n" );
      fprintf( outf, "         sub.l    d0,mc_val\n" );
      break;

   case '*':

      fprintf( outf, "         move.l   #%d,d1\n", (reladr<<2) );
      fprintf( outf, "         bsr      GetVar\n" );
      fprintf( outf, "         move.l   mc_val(pc),d1\n" );
      fprintf( outf, "         muls.l   d0,d1\n" );
      fprintf( outf, "         move.l   d1,mc_val\n" );
      break;

   case '/':
      fprintf( outf, "         move.l   #%d,d1\n", (reladr<<2) );
      fprintf( outf, "         bsr      GetVar\n" );
      fprintf( outf, "         move.l   mc_val(pc),d1\n" );
      fprintf( outf, "         divs.l   d0,d1\n" );
      fprintf( outf, "         move.l   d1,mc_val\n" );
      break;


   case '>':
//    mc->SetData( correctblock+reladr, mc->val );
      fprintf( outf, "         move.l   Stack(pc),a1\n" );
      fprintf( outf, "         add.l    d0,a1\n" );
      fprintf( outf, "         add.l    #%d,a1\n", (reladr<<2) );
      fprintf( outf, "         move.l   mc_val(pc),(a1)\n" );
      break;

   case '?':
      fprintf( outf, "         move.l   d0,-(a7)\n" );

      fprintf( outf, "         move.l   #InBufAsk,d1\n" );
      fprintf( outf, "         move.l   _DosBase(pc),a6\n" );
      fprintf( outf, "         jsr      _LVOPutStr(a6)\n" );
      fprintf( outf, "         move.l   _Output(pc),d1\n" );
      fprintf( outf, "         jsr      _LVOFlush(a6)\n" );


      fprintf( outf, "         move.l   _Input(pc),d1\n" );
      fprintf( outf, "         move.l   _DosBase(pc),a6\n" );
      fprintf( outf, "         move.l   #InBuf,d2\n" );
      fprintf( outf, "         move.l   #11,d3\n" );
      fprintf( outf, "         jsr      _LVORead(a6)\n" );

      fprintf( outf, "         moveq    #0,d0\n" );
      fprintf( outf, "         lea      InBuf(pc),a1\n" );
      fprintf( outf, "         move.b   (a1),d0\n" );
      fprintf( outf, "         sub.b    #'0',d0\n" );

      fprintf( outf, "         move.l   (a7)+,d1\n" );

      fprintf( outf, "         move.l   Stack(pc),a1\n" );
      fprintf( outf, "         add.l    d1,a1\n" );
      fprintf( outf, "         add.l    #%d,a1\n", (reladr<<2) );
      fprintf( outf, "         move.l   d0,(a1)\n" );


//      printf( "Input for variable %s: ", name );
//      scanf( "%d", &value );
//      mc->SetData( correctblock+reladr, value );
      break;

   case '!':
      fprintf( outf, "         move.l   #%d,d1\n", (reladr<<2) );
      fprintf( outf, "         bsr      GetVar\n" );
      fprintf( outf, "         move.l   d0,OutArg\n" );
      fprintf( outf, "         lea      OutString(pc),a0\n" );
      fprintf( outf, "         lea      OutArg(pc),a1\n" );
      fprintf( outf, "         lea      OutChar(pc),a2\n" );
      fprintf( outf, "         lea      OutBuf(pc),a3\n" );
      fprintf( outf, "         move.l   _ExecBase(pc),a6\n" );
      fprintf( outf, "         jsr      _LVORawDoFmt(a6)\n" );

      fprintf( outf, "         move.l   #OutBuf,d1\n" );
      fprintf( outf, "         move.l   _DosBase(pc),a6\n" );
      fprintf( outf, "         jsr      _LVOPutStr(a6)\n" );

//      printf( "Output for variable %s: %d\n", name, value );
      break;

   }


   return 1;
}

void Var::Dump( void )
{
   printf( "Var(%s)\n", name );
}



Operator::Operator( Machine *emc, char eop )
{
   mc = emc;

   operation = eop;

}

int Operator::Execute( void )
{
   mc->op = operation;

   return 1;
}

int Operator::GenCode( FILE *outf )
{
   // This is a dummy procedure when executing,
   // since nilco-kludge-o-matics found out the operator in previous instruction

   return 1;
}

void Operator::Dump( void )
{
   printf( "Operator(%c)\n", operation );
}


ThenAction::ThenAction( Machine *emc )
{
   mc = emc;
}

int ThenAction::Execute( void )
{
   if( mc->val >= 0 ) ;

//      printf( "Then(nojump,ip=%d)\n", mc->GetIP() );

   else {
//      printf( "Then(jump,ip=%d)\n", JumpTo );
      mc->SetIP( JumpTo );
   }

   return 1;
}

int ThenAction::GenCode( FILE *outf )
{
   fprintf( outf, "         bmi      .llab_%d\n", JumpTo );
   return 1;
}

void ThenAction::Dump( void )
{
   printf( "Then(ip=%d)\n", JumpTo );
}


EndAction::EndAction( Machine *emc )
{
   mc = emc;
}

int EndAction::Execute( void )
{
   int popsize;

//   printf( "End(%d)\n", mc->GetData( mc->GetCur() + OFF_RetAddr ) );

   mc->SetIP( mc->GetData( mc->GetCur() + OFF_RetAddr ) );
   popsize = mc->GetFree() - mc->GetCur();
   mc->SetCur( mc->GetData( mc->GetCur() + OFF_DynamicLink ) );    // call??

   if( !mc->Pop( popsize ) ) return 0;

   return 1;
}

int EndAction::GenCode( FILE *outf )
{
   fprintf( outf, "         bsr      ClearCall\n" );
   fprintf( outf, "         rts\n" );
   return 1;
}

void EndAction::Dump( void )
{
   printf( "End\n" );
}


Stop::Stop( Machine *emc )
{
   mc = emc;
}

int Stop::Execute( void )
{
   puts( "*** Execution stopped" );
   mc->SetIP( -1 );

   return 1;
}

int Stop::GenCode( FILE *outf )
{
   fprintf( outf, "         bra      Cleanup\n" );
   return 1;
}

void Stop::Dump( void )
{
   printf( "Stop\n" );
}

