/****
 * Name: logla.cpp
 *
 * Details:
 * The main program.
 *
 * Bugs:
 ****/

#include "logla.h"

char *version = "$VER: Logla-1 0.2 (7.1.96)";

void Compile( struct CompileObj *co )
{
   Reader Read;
   Machine Mach( 1000, 1000, 1000 );
   Parser *Parse;
   int ret;

   if( Read.LoadFile( co->fname ) ) {

      puts( "*** Parsing & creating pseudo code..." );

      Parse = new Parser( &Read, &Mach, co->pdb );

      if( Parse->Program() ) {

         puts( "*** Parsing ok!" );
         if( co->mdb_dump ) Mach.DumpInst();

         if( !co->no68k ) ret = Mach.Generate( co->srcname );

         if( !ret ) {

            if( !co->nocomp && !co->no68k ) ret = Mach.GenIt( co->asmstring, co->srcname );

            if( !ret )
               if( co->m_run ) Mach.Interpreter( co->mdb_ss );

         }

      } else
         puts( "Error parsing! Compilation stopped!" );

   } else
      puts( "Error reading file! Compilation stopped!" );


}


int main( int argc, char *argv[] )
{
   struct CompileObj co = { 0 };

   int i;

   int argpos;
   int arglen;

   puts( "\nLogla-1 Compiler v0.2 (7.1.96) by Nils Corneliusen of TRIUMPH Software" );
   puts( "----------------------------------------------------------------------" );

   if( argc == 1 ) {
      puts( "Usage: logla [-pmsrcng] <source file>" );
      puts( "Type logla -h for help." );
      return 0;
   }

   if( !strcmp( argv[1], "-h" ) ) {
      puts( "Usage: LOGLA [-pmsrcng] <source file>" );
      puts( "       -p: Show parsing" );
      puts( "       -m: Dump pseudo-code after parsing" );
      puts( "       -s: Use pseudo-code debugger if -r specified" );
      puts( "       -r: Run/interpret pseudo-code" );
      puts( "       -c: Do not generate 68k assembler code" );
      puts( "       -n: Do not compile 68k code" );
      puts( "       -g: Specify assembler. Default is \"genam %s\"" );
      puts( "           Enclose this option with \", insert %s for codename." );
      puts( "           Example: logla \"-ggenam -oram:test %s\" dir.logla" );
      return 0;
   }

   i = 1;

   while( i <= argc - 2 ) {

      if( !(argv[i][0] == '-') ) {
         puts( "Expecting arguments!" );
         return 0;
      }

      argpos = 1;
      arglen = ((int)strlen( argv[i] ));

      while( argpos != arglen ) {

         switch( argv[i][argpos] ) {

         case 'p': co.pdb = 1;      break;
         case 'm': co.mdb_dump = 1; break;
         case 's': co.mdb_ss = 1;   break;
         case 'r': co.m_run = 1;    break;
         case 'c': co.no68k = 1;    break;
         case 'n': co.nocomp = 1;   break;
         case 'g':
            strncpy( co.asmstring, argv[i]+argpos+1, 255 );
            argpos = arglen - 1;
            break;


         default: printf( "Unknown argument '%c'!\n", argv[i][argpos] ); return 0; break;

         }

         argpos++;

      }

      i++;

   }

   strcpy( co.asmstring, "genam %s" );
   strcpy( co.fname, argv[argc-1] );

   arglen = ((int)strlen( argv[argc-1] ));

   if( arglen < 7 )
      sprintf( co.srcname, "%s.s", argv[argc-1] );
   else {
      if( strcmp( argv[argc-1] + arglen - 6, ".logla" ) )
         sprintf( co.srcname, "%s.s", argv[argc-1] );
      else {
         strncpy( co.srcname, argv[argc-1], arglen-6 );
         co.srcname[arglen-6] = 0;
         strcat( co.srcname, ".s" );
      }
   }


   Compile( &co );

   return 0;

}

