/****
 * Name: machine.cpp
 *
 * Details:
 * The virtual machine.
 * Mainly used by the interpreter. Contains all normal machine stuff, like
 * the debugger etc. Also takes care of code generation if needed.
 *
 * Bugs:
 ****/


#include <stdio.h>
#include <stdlib.h>

#include "machine.h"


Machine::Machine( int emp, int emi, int emd )
{
   nextIA = 0;
   nextPD = 0;

   maxproc = emp;
   maxinst = emi;
   maxdata = emd;

   IA = new (Inst (*[maxinst]));
   PD = new (ProcDesc (*[maxproc]));
   D  = new int[maxdata];

   cur = 0;

   ip = 0;

   // generate fake stackframe

   D[2] = -1;
   free = 3;
   bll = 0;

}


void Machine::Interpreter( int single_step )
{
   int CurIP;

   puts( "*** Running interpreter..." );

   if( single_step ) {
      puts( "Single step debugging commands:" );
      puts( "RETURN = execute command" );
      puts( "R      = run" );
      puts( "D      = dump memory" );
   }

   while( ip != -1 ) {

      CurIP = ip;

      ip++;

      if( single_step ) {
         printf( "IA[%4d] ", CurIP );
         IA[CurIP]->Dump();
         getchar();
      }

      if( !IA[CurIP]->Execute() ) ip = -1;

      // debug??

   }


}



void Machine::DumpInst( void )
{
   puts( "*** Dumping instructions..." );

   for( int i = 0; i < nextIA; i++ ) {

      printf( "IA[%4d] ", i );
      IA[i]->Dump();

   }


}


int Machine::Generate( char *srcfile )
{
   FILE *outf;

   if( outf = fopen( srcfile, "w" ) ) {

      puts( "*** Generating code..." );

      currIA = 0;

      for( int i = 0; i < nextIA; i++ ) {

         IA[i]->GenCode( outf );
         currIA++;

      }

      fclose( outf );

   } else
      printf( "*** Error generating source file '%s'!", srcfile );

   return outf==0;
}


int Machine::GenIt( char *asmstring, char *srcfile )
{
   int ret;
   char buffer[200];

   puts( "*** Compiling code..." );

   sprintf( buffer, asmstring, srcfile );

   ret = system( buffer );

   if( ret ) puts( "*** Compilation failed!" );

   return ret;

}



int Machine::CreateInst( Inst *instr )
{
   if( nextIA >= maxinst ) {
      puts( "Machine/CreateInst: Instruction (IA) storage full!" );
      printf( "Limit = %d, use HEY option to set it higher.\n", maxinst );
      return 0;
   }

   IA[nextIA] = instr;
   nextIA++;
   return 1;

}

int Machine::CreateProc( ProcDesc *proc )
{
   if( nextPD >= maxproc ) {
      puts( "Machine/CreateProc: Procedure (PD) storage full!" );
      printf( "Limit = %d, use HEY option to set it higher.\n", maxproc );
      return 0;
   }


   PD[nextPD] = proc;
   nextPD++;
   return 1;

}

int Machine::GetIP( void )
{

   return ip;

}

void Machine::SetIP( int newip )
{
   ip = newip;
}

int Machine::GetData( int offset )
{
   return D[offset];
}

int Machine::SetData( int offset, int val )
{
   D[offset] = val;
   return 1;
}

int Machine::Push( int size )
{
   if( free + size > maxdata ) {
      puts( "Machine/Push: Data (D) storage full!" );
      printf( "Limit = %d, use HEY option to set it higher.\n", maxdata );
      return 0;
   }

   cur = free;
   free = free + size;
   return 1;

}

int Machine::Pop( int size )
{

   if( free - size < 0 ) {
      puts( "Machine/Pop: Data (D) storage already empty!" );
      puts( "Maybe somebody generated fuzzy code? :-)" );
      return 0;
   }

   free = free - size;
   return 1;
}

int Machine::GetCur( void )
{
   return cur;
}

void Machine::SetCur( int ncur )
{
   cur = ncur;
}

// get "pointer" to top of stack
int Machine::GetFree( void )
{
   return free;
}

