/****
 * Name: parser.cpp
 *
 * Details:
 * The parser.
 * Parses the source file. Calls Reader class functions to get symbols.
 * Uses a normal recursive descent method for LL(1) syntax.
 *
 * Bugs:
 * - does not build a parse tree, since this is not needed. Maybe add for fun? :-)
 *   (this might be smart for later optimization passes or symbol backpatching)
 * - fishy *Name calls. Should fix this...
 ****/


#include <stdio.h>
#include <string.h>

#include "parser.h"
#include "decl.h"
#include "procdesc.h"
#include "inst.h"

Parser::Parser( Reader *rl, Machine *emc, int pd )
{

   pdebug = pd;
   l = rl;
   mc = emc;
   cm = new Compiler( mc, 1000 );

}


char *Parser::AName( char *dest )
{
   if( pdebug ) l->ViewMsg( "AName" );

   if( isdigit( (l->LastSym())[0] ) ) {
      l->ViewErr( "AName: Illegal name!" );
      return 0;
   }

//   if( !l->NextSym() ) return 0;

   return l->LastSym();
}

Decl *Parser::PName( void )
{
   Decl *nproc;

   if( pdebug ) l->ViewMsg( "PName" );

   if( isdigit( (l->LastSym())[0] ) ) {
      l->ViewErr( "PName: Illegal procedure name!" );
      return 0;
   }

   nproc = new Decl( l->LastSym(), TYPE_PROC, mc->nextPD, cm->blocklevel );

   if( !cm->NewProc( nproc ) ) {
      l->ViewErr( "PName: Error defining new procedure!" );
      return 0;
   }

   if( !l->NextSym() ) return 0;

   return nproc;

}

int Parser::NName( void )
{
   if( pdebug ) l->ViewMsg( "NName" );

   if( isdigit( l->LastSym()[0] ) ) {
      l->ViewErr( "NName: Illegal variable name!" );
      return 0;
   }

   if( !cm->NewVar( l->LastSym() ) ) {
      l->ViewErr( "NName: Error defining new variable!" );
      return 0;
   }

   return l->NextSym();

}


int Parser::XNameList( void )
{

   if( pdebug ) l->ViewMsg( "XNameList" );

   if( !strcmp( l->LastSym(), "," ) ) {

      if( !l->NextSym() ) return 0;
      if( !NName() ) return 0;
      return XNameList();

   }

   if( strcmp( l->LastSym(), ";" ) ) {

      l->ViewErr( "XNameList: Expecting ';'!" );
      return 0;

   }

   return 1;

}

int Parser::NameList( void )
{
   if( pdebug ) l->ViewMsg( "NameList" );

   if( !NName() ) return 0;

   return XNameList();

}

int Parser::VarDecl( void )
{
   if( pdebug ) l->ViewMsg( "VarDecl" );

   if( strcmp( l->LastSym(), "VAR" ) ) {
      l->ViewErr( "VarDecl: Expecting VAR!" );
      return 0;
   }

   if( !l->NextSym() ) return 0;

   return NameList();
}



int Parser::Decl2( void )
{
   if( pdebug ) l->ViewMsg( "Decl2" );

        if( !strcmp( l->LastSym(), "PROCEDURE" ) ) return ProcDecl();
   else if( !strcmp( l->LastSym(), "VAR" ) ) return VarDecl();

   l->ViewErr( "Decl2: Expecting PROCEDURE or VAR!" );
   return 0;

}

int Parser::DeclList( void )
{
   if( pdebug ) l->ViewMsg( "DeclList" );

   if( !strcmp( l->LastSym(), "BEGIN" ) ) return 1;

   if( strcmp( l->LastSym(), "PROCEDURE" ) && strcmp( l->LastSym(), "VAR" ) ) {
      l->ViewErr( "DeclList: Expecting PROCEDURE or VAR!" );
      return 0;
   }

   if( !Decl2() ) return 0;

   if( strcmp( l->LastSym(), ";" ) ) {
      l->ViewErr( "DeclList: ';' expected at end of declarations!" );
      return 0;
   }

   if( !l->NextSym() ) return 0;

   return DeclList();

}



int Parser::Variable( void )
{
   char destvar[80];
   Decl *ddecl;

   if( pdebug ) l->ViewMsg( "Variable" );

   strcpy( destvar, l->LastSym() );

   if( !l->NextSym() ) return 0;

   if( !isalpha( (int) destvar[0] ) ) {
      l->ViewErr( "Variable: Expecting variable name!" );
      return 0;
   }

   if( !(ddecl = cm->FindDecl( destvar )) ) {
      l->ViewErr( "Variable: Variable does not exist!" );
      return 0;
   }

   if( ddecl->type == TYPE_PROC ) {
      l->ViewErr( "Variable: Trying to access procedure as variable!" );
      return 0;
   }

   if( !mc->CreateInst( new Var( mc, cm->blocklevel-ddecl->bllevel, ddecl->nr, destvar ) ) ) return 0;

   return 1;

}

int Parser::Number( void )
{

   if( pdebug ) l->ViewMsg( "Number" );

   if( !mc->CreateInst( new FNumber( mc, atoi( l->LastSym() ) ) ) ) return 0;

   if( !l->NextSym() ) return 0;

   return 1;
}


int Parser::Term( void )
{
   char *tmp;

   if( pdebug ) l->ViewMsg( "Term" );

   tmp = l->LastSym();

   if( isdigit( tmp[0] ) ) return Number();

   return Variable();

}

int Parser::Arop( void )
{
   if( pdebug ) l->ViewMsg( "Arop" );

   if( !strcmp( l->LastSym(), ">" ) || !strcmp( l->LastSym(), "THEN" ) ) return 1;

   if( strcmp( l->LastSym(), "+" ) && strcmp( l->LastSym(), "-" ) &&
       strcmp( l->LastSym(), "*" ) && strcmp( l->LastSym(), "/" ) ) {

      l->ViewErr( "Arop: Expecting '+' '-' '*' '/'!" );
      return 0;
   }

   if( !mc->CreateInst( new Operator( mc, l->LastSym()[0] ) ) ) return 0;

   if( !l->NextSym() ) return 0;

   return 1;
}




int Parser::XExpression( void )
{
   if( pdebug ) l->ViewMsg( "XExpression" );

   if( !strcmp( l->LastSym(), ">" ) || !strcmp( l->LastSym(), "THEN" ) ) return 1;

   if( strcmp( l->LastSym(), "+" ) && strcmp( l->LastSym(), "-" ) &&
       strcmp( l->LastSym(), "*" ) && strcmp( l->LastSym(), "/" ) ) {

      l->ViewErr( "XExpression: Expecting '+' '-' '*' '/' or THEN '>'!" );
      return 0;
   }

   if( !Arop() ) return 0;
   if( !Term() ) return 0;

   return XExpression();

}


int Parser::Expression( void )
{
   if( pdebug ) l->ViewMsg( "Expression" );

   if( !Term() ) return 0;

   return XExpression();

}



int Parser::IfStmt( void )
{
   int place;

   if( pdebug ) l->ViewMsg( "IfStmt" );

   if( strcmp( l->LastSym(), "IF" ) ) {
      l->ViewErr( "If: Expecting IF!" );
      return 0;
   }

   if( !l->NextSym() ) return 0;

   if( !mc->CreateInst( new Operator( mc, 'I' ) ) ) return 0;

   if( !Expression() ) return 0;

   if( strcmp( l->LastSym(), "THEN" ) ) {
      l->ViewErr( "If: Expecting THEN!" );
      return 0;
   }

   if( !l->NextSym() ) return 0;

   // *** thenaction uses JumpTo to resolve local jump

   place = mc->nextIA;
   if( !mc->CreateInst( new ThenAction( mc ) ) ) return 0;



   l->In();
   if( !StmtList() ) return 0;
   l->Out();


   // call?
   ((ThenAction *)mc->IA[place])->JumpTo = mc->nextIA;

   if( !mc->CreateInst( new LLabel( mc, mc->nextIA ) ) ) return 0;

   if( strcmp( l->LastSym(), "FI" ) ) {
      l->ViewErr( "If: Expecting FI!" );
      return 0;
   }

   return l->NextSym();
}


int Parser::Output( void )
{
   if( pdebug ) l->ViewMsg( "Output" );

   if( strcmp( l->LastSym(), "OUTPUT" ) ) {
      l->ViewErr( "Output: Expecting OUTPUT!" );
      return 0;
   }

   if( !l->NextSym() ) return 0;

   if( !mc->CreateInst( new Operator( mc, '!' ) ) ) return 0;

   return Variable();
}


int Parser::Input( void )
{
   if( pdebug ) l->ViewMsg( "Input" );

   if( strcmp( l->LastSym(), "INPUT" ) ) {
      l->ViewErr( "Input: Expecting INPUT!" );
      return 0;
   }

   if( !l->NextSym() ) return 0;

   if( !mc->CreateInst( new Operator( mc, '?' ) ) ) return 0;

   return Variable();
}



int Parser::Assignment( void )
{
   if( pdebug ) l->ViewMsg( "Assignment" );

   if( strcmp( l->LastSym(), "ASSIGN" ) ) {
      l->ViewErr( "Assignment: Expecting ASSIGN!" );
      return 0;
   }

   if( !l->NextSym() ) return 0;


   if( !mc->CreateInst( new Operator( mc, 'A' ) ) ) return 0;

   if( !Expression() ) return 0;

   if( strcmp( l->LastSym(), ">" ) ) {
      l->ViewErr( "Assignment: Expecting '>'!" );
      return 0;
   }

   if( !l->NextSym() ) return 0;

   if( !mc->CreateInst( new Operator( mc, '>' ) ) ) return 0;

   return Variable();
}


int Parser::ProcName( void )
{
   char proc[80];
   char *tmp;
   Decl *pr2call;

   if( pdebug ) l->ViewMsg( "ProcName" );

   if( !(tmp = AName( proc )) ) return 0;
   strcpy( proc, tmp );
   if( !l->NextSym() ) return 0;

   if( !(pr2call = cm->FindDecl( proc )) ) {
      l->ViewErr( "ProcName: Procedure does not exist!" );
      return 0;
   }

   if( pr2call->type == TYPE_VAR ) {
      l->ViewErr( "ProcName: You cannot call a variable!" );
      return 0;
   }

   if( !mc->CreateInst( new FCall( mc, pr2call->nr ) ) ) return 0;


   // special debug


   return 1;
}


int Parser::Call( void )
{
   if( pdebug ) l->ViewMsg( "Call" );

   if( strcmp( l->LastSym(), "CALL" ) ) {
      l->ViewErr( "Call: Expecting CALL!" );
      return 0;
   }

   if( !l->NextSym() ) return 0;

   return ProcName();

}


int Parser::Stmt( void )
{
   if( pdebug ) l->ViewMsg( "Stmt" );

        if( !strcmp( l->LastSym(), "CALL"   ) ) return Call();
   else if( !strcmp( l->LastSym(), "ASSIGN" ) ) return Assignment();
   else if( !strcmp( l->LastSym(), "INPUT"  ) ) return Input();
   else if( !strcmp( l->LastSym(), "OUTPUT" ) ) return Output();
   else if( !strcmp( l->LastSym(), "IF"     ) ) return IfStmt();

   l->ViewErr( "Stmt: Expecting CALL/ASSIGN/INPUT/OUTPUT/IF!" );
   return 0;

}



int Parser::StmtList( void )
{
   if( pdebug ) l->ViewMsg( "StmtList" );

   if( !strcmp( l->LastSym(), "FI" ) || !strcmp( l->LastSym(), "END" ) ) return 1;

   if( !strcmp( l->LastSym(), "ASSIGN" ) || !strcmp( l->LastSym(), "IF" ) ||
       !strcmp( l->LastSym(), "CALL" )   || !strcmp( l->LastSym(), "OUTPUT" ) ||
       !strcmp( l->LastSym(), "INPUT" ) ) {

          if( !Stmt() ) return 0;

   } else {
      l->ViewErr( "StmtList: Expecting ASSIGN/IF/CALL/OUTPUT/INPUT or END/FI!" );
      return 0;
   }

   if( strcmp( l->LastSym(), ";" ) ) {
      l->ViewErr( "StmtList: Expecting ';' at end of statement!" );
      return 0;
   }

   if( !l->NextSym() ) return 0;

   return StmtList();

}


int Parser::ProcDecl( void )
{
   Decl *tproc;


   if( pdebug ) l->ViewMsg( "ProcDecl" );

   l->In();

   if( strcmp( l->LastSym(), "PROCEDURE" ) ) {
      l->ViewErr( "ProcDecl: PROCEDURE expected!" );
      return 0;
   }

   if( !l->NextSym() ) return 0;

   // hmm
   if( !(tproc = PName()) ) return 0;

   if( !l->NextSym() ) return 0;

   if( !DeclList() ) return 0;

   // state explicitly where label is to be generated
   if( !mc->CreateInst( new GLabel( mc, tproc->nr ) ) ) return 0;

   if( strcmp( l->LastSym(), "BEGIN" ) ) {
      l->ViewErr( "ProcDecl: BEGIN expected!" );
      return 0;
   }

   if( !l->NextSym() ) return 0;


   //should be a call
   mc->PD[tproc->nr]->startinst = mc->nextIA;

   l->In();
   if( !StmtList() ) return 0;
   l->Out();


   if( strcmp( l->LastSym(), "END" ) ) {
      l->ViewErr( "ProcDecl: END expected!" );
      return 0;
   }

   if( !l->NextSym() ) return 0;

   l->Out();

   if( !mc->CreateInst( new EndAction( mc ) )) return 0;

   cm->LeaveProc();

   return 1;

}


int Parser::Program( void )
{

   if( pdebug ) l->ViewMsg( "Program" );

   if( !mc->CreateInst( new FProgram( mc ) ) ) return 0;

   if( !mc->CreateInst( new FCall( mc, 0 ) ) ) return 0;
   if( !mc->CreateInst( new Stop( mc ) ) ) return 0;

   if( !l->NextSym() ) return 0;

   if( !ProcDecl() ) return 0;

   return mc->CreateInst( new Defs( mc ) );

}


